;;; -*- auto-recompile: t -*-

;;; csound-lsp.el --- Csound composition in Emacs Lisp

;; Keywords: csound

;; This file is not part of GNU Emacs.
;; 
;; csound-lsp.el is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; csound-lsp.el is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Author: Stphane Rollandin <hepta@zogotounga.net>
;;---------

;;; Commentary:
;;-------------

;;    full documentation here:
;;    http://www.zogotounga.net/comp/csoundx.html

;;
;;; Installation:
;;                  
;; ==========================================================
;; this file should be installed through the csound-x package
;; ==========================================================
;;

;; last modified April 10, 2006

;;; Code:

(require 'csound-csd)

(defgroup csound-lsp nil
  "Emacs Lisp for Csound composition"
  :group 'csound-x
  :prefix "csl-")

(defmacro with-csl-score-template (&rest body)
  `(progn
     (insert "<CsScore>\n")
     ,@body
     (insert "\n</CsScore>\n")))

(defmacro with-csl-orchestra-template (&rest body)
  `(progn
     (setq csl-varname-counter 1000) 
     (insert "<CsInstruments>
sr        =         44100
kr        =         4410
ksmps     =         10
nchnls    =         1
" ?\n)
     ,@body
     (insert "\n</CsInstruments>\n")))

(defvar csl-varname-counter 1000)
(make-variable-buffer-local 'csl-varname-counter)

(defun csl-new-varname (&optional opcode)
  (format "%s%s" (if opcode (substring (concat opcode "vv") 0 3)
                   "var")
          (incf csl-varname-counter)))

; (csl-new-varname "linen")
; (csl-new-varname "oscili")
; (csl-new-varname "in")

;==========

(defun csl-transcribe-to-code (sexp)
  ""
  (let* ((opcode (symbol-name (first sexp)))
         (fopcodep (member opcode csdoc-functions))
         output opargs code-lines globalp)
    (dolist (arg (cdr sexp))
      (cond
       ;; single keyword (:arate, :krate or :irate)
       ((keywordp arg)
        (if (eq arg :global)
            (setq globalp t)
          (setq output (csl-output-for opcode arg))))
       ;; list
       ((listp arg)
        ;; output spec (:output k1 a2 ...)
        (if (equal (first arg) :output)       
            (setq output (mapcar 'symbol-name (cdr arg)))
          ;; else another sexp to expand
          (let ((rec (csl-transcribe-to-code arg)))
            (add-to-list 'opargs (caar rec) t)
            (setq code-lines (append code-lines (cadr rec))))))
       ;; a plain symbol or number 
       (t
        (add-to-list 'opargs (format "%s" arg) t))))
    (setq output (or output (if fopcodep
                                (apply 'csl-f-output-from opargs)
                              (csl-output-for opcode))))
    (when globalp
      (setq output (mapcar (lambda (o) (unless (eq (elt o 0) ?g) (concat "g" o))) output)))
    (add-to-list 'code-lines
                 (if fopcodep
                     (format "%s = %s" 
                             (car output)
                             (if (null (cdr opargs))
                                 (concat opcode "(" (first opargs) ")")
                               (mapconcat 'identity opargs opcode)))
                   (format "%s %s %s"
                           (mapconcat 'identity output ", ") 
                           opcode
                           (mapconcat 'identity opargs ", "))) 
                 t)
    (list output code-lines)))

;==========


(defun csl-output-for (opcode &optional rate global)
  (unless (or (member opcode csdoc-0-opcodes)
              (member opcode csdoc-functions))
    (let ((spec (car (split-string (csdoc-opcode :spec opcode))))
          (output))
      (dotimes (o (length spec))
        (add-to-list 'output        
                     (concat (string (if rate
                                         (elt (symbol-name rate) 1) 
                                       (elt spec o)))
                             (csl-new-varname opcode))
                     t))
      output)))

; (csl-output-for "linen" :arate t)
; (csl-output-for "linen" :krate)
; (csl-output-for "linen")

(defun csl-f-output-from (var1 &optional var2)      
  (list (concat
         (if (null var2)
             (elt var1 0) (csl-new-varname) 
             (let ((spec (list (elt var1 0) (elt var2 0))))
               (cond ((member ?a spec) "a")
                     ((member ?k spec) "k")
                     (t "i"))))
         (csl-new-varname))))

; (csl-f-output-from "ahoui" "kbof")


; (csl-transcribe-to-code '(out (oscili :arate (linen :krate p4 p6 p3 p7) p5 p8)))

(defun csl-insert-code (quoted-sexp)
  (let ((code (csl-transcribe-to-code quoted-sexp)))
    (dolist (line (cadr code))
      (insert line "\n"))
    (car code)))

(defun csl-insert-instrument (name &rest body)
  (insert (format "instr %s\n" name))
  (dolist (code body)
    (csl-insert-code code))
  (insert "endin\n\n"))

(put 'csl-insert-instrument 'lisp-indent-function 1)

(defun csl-play-composition (composition)
  (with-csound-to-DAC (funcall composition)))

(defun csl-render-composition (composition wavfile)
  (with-csound-to-WAV wavfile (funcall composition)))

(defcustom csl-composition-buffer-name "Csound composition"
  "Name for buffers created for editing a csound composition"
  :type 'string
  :group 'csound-lsp)

(defun csl-edit-composition (composition)
  (switch-to-buffer (generate-new-buffer csl-composition-buffer-name))
  (funcall composition)
  (goto-char (point-min))
  (csound-csd-mode))


;; ===================================================================
;;                   f-tables utilities
;; ===================================================================

(defcustom csl-tables '(:sin "f %s 0 %s 10 1")
  "Templates for GEN tables
First replacement slot is for table number, second for table size"
  :type 'plist
  :group 'csound-lsp)

(defun csl-insert-ftable (num type size)
  (insert (format (plist-get csl-tables type) num (expt 2 (round (log size 2)))) ?\n))


;; ===================================================================
;;                   the 'csound-composition macro
;; ===================================================================

'(example 
  (csl-edit-composition
;  (csl-play-composition  
   (csound-composition
    :ftables '(1 :sin 8192)
    :instr 1
    '(out (oscili :arate (oscili :krate p4 1/p3 p7) p5 p6))
    :score
    (insert "f 2 0 513 5 1 12 1024 500 1" ?\n)
    (i 1 0 4 8000 440 1 2)
    (loop for h in '(2200 600 215 1852 990)
          for d in '(2 2.5 2 1.5 4)
          do (+i :p5 h :p3 d))))
)

(defmacro csound-composition (&rest args)
  "eval ARGS in the context of a new blank CSD buffer
keywords delimiters define the working area in the CSD:
:ftables 
:instr
:score
:options
:globals
"
  (let ((cscompo-orc-code nil)
        (cscompo-sco-statements nil))
    (dolist (chunk 
             (loop for form in args
                   for fplace upfrom 1
                   if (keywordp form)
                   collect `(,form ,@(loop for f in (nthcdr fplace args)
                                           until (keywordp f) 
                                           collect f))))
      (case (car chunk)
        (:ftables     
         (mapcar (lambda (fspec) (push `(apply 'csl-insert-ftable ,fspec) cscompo-sco-statements)) 
                 (cdr chunk)))
        (:instr
         (add-to-list 'cscompo-orc-code
                      `(csl-insert-instrument ,(cadr chunk) ,@(cddr chunk))
                      t))
        (:opcode ())
        (:score
         (nconc cscompo-sco-statements (cdr chunk)))
        (:i-stream          
         (nconc cscompo-sco-statements `((with-pfields-stream ,@(cdr chunk)))))
        (:options ())
        (:globals ())))
    `(lambda ()
       (with-csl-orchestra-template
        ,@cscompo-orc-code)
       (with-csl-score-template
        ,@cscompo-sco-statements))))


;; ===================================================================
;;                   testing
;; ===================================================================

(defvar csl-example-composition 
  (csound-composition
   :instr 1108
   '(out (oscili :arate (linen :krate p4 p6 p3 p7) p5 p8))
   :instr 1109
   '(out (* (:output aout) 
            (linen (:output kline) p4 p6 p3 p7)
            (+ (oscili (:output aosc1) 1 p5 p9)
               (oscili (:output aosc2) 1 p8 p9))))
   :ftables '(1 :sin 8192)
   :score
   (i 1108 0 2 15000 400 .3 .3 1)
   (i 1108 3 2 15000 10 .3 .3 1)
   (i 1109 6 2 25000 400 .3 .3 10 1)
   ;; (s)
   (insert "s\n")
   (i 1108 1 2 15000 400 .3 .3 1)
   (i 1108 4 2 15000 170 .3 .3 1)
   (i 1109 7 2 25000 400 .3 .3 170 1)  
   ;; (s)
   (insert "s\n")
   (i 1108 1 2 15000 400 .3 .3 1)
   (i 1108 4 2 15000 385 .3 .3 1)
   (i 1109 7 2 25000 400 .3 .3 385 1)))

(defun csl-test-dac ()
  (interactive)
  (csl-play-composition csl-example-composition))

(defun csl-test-wav ()
  (interactive)
  (csl-render-composition csl-example-composition "test.wav"))

;; === this is it.
(provide 'csound-lsp)

;; csound-lsp.el ends here

