;;; -*- auto-recompile: t -*-

;;; tests.el --- simple utilities for testing elisp code

;; This file is not part of GNU Emacs.
;; 
;; tests.el is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; tests.el is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


;; Installation:
;;-------------
;;
;;         (require 'tests)
;;
;; Usage:
;;-------
;;
;;  in emacs lisp code, use comment lines such as the following to 
;; set up a test:

; TEST (concat "this " "is" " a test") => "this is a test"

;; those lines are automatically recognized. use the "Tests" menu 
;; to perform one or all tests.
;;
;; if you have embedded-elisp-library.el installed, the menu
;; also allows you to have a button for each test.
;;
;;
;; author: Stphane Rollandin <hepta@zogotounga.net>
;; last modified April 11, 2006

(require 'cl)
(require 'lisp-mode)
(require 'easymenu)

;; ===================================================================
;;                 core functions
;; ===================================================================

;; (deftest '(= (1+ 1) 2) 'test-examples)
;; (deftest '(= (1+ 1) 3) 'test-examples)
;; (runtests 'test-examples)

(defvar tests-list ())
(defvar tests-output-buffer "*Tests*")

(defun deftest (quoted-form header &optional list)
  (add-to-list (or list 'tests-list)
	       (list header quoted-form)
	       t))

(defun runtests (header &optional list)
  (get-buffer-create tests-output-buffer)
  (let ((nerrors 0) (ntests 0))
    (dolist (test (or list tests-list))
      (when (equalp (car test) header)
	(incf ntests)
	(unless (eval (cadr test))
	  (incf nerrors)
	  (signal-test-failure test))))
    (save-excursion
      (set-buffer tests-output-buffer)
      (goto-char (point-max))
      (let ((conclusion (format "%d test(s) performed. %d failure(s)" ntests nerrors)))
	(insert conclusion "\n\n\n")
	(message conclusion)))
    (unless (zerop nerrors)
      (switch-to-buffer-other-window tests-output-buffer))))

(defun signal-test-failure (atest)
  (save-excursion
    (set-buffer tests-output-buffer)
      (goto-char (point-max))
    (insert "Failure:\n" (prin1-to-string (cadr atest)) "\n\n")))


;; ===================================================================
;;                 comment-lines tests
;; ===================================================================

;TEST (1+ 1) => 2
;TEST (1+ 2) => 2

(defun go-to-next-commented-test (&optional upto)
  (interactive)
  (or (re-search-forward "^[; \t]*TEST\\(.*\\)=>\\(.*\\)" upto t)
      (when (and (interactive-p)
                 (y-or-n-p "No more tests. Scan again from the beginning ?"))
        (goto-char (point-min))
        (go-to-next-commented-test))))

(defun go-to-previous-commented-test ()
  (interactive)
  (or (re-search-backward "^[; \t]*TEST\\(.*\\)=>\\(.*\\)" nil t)
      (when (and (interactive-p)
                 (y-or-n-p "No more tests. Scan again from the end ?"))
        (goto-char (point-max))
        (go-to-previous-commented-test))))

(defun tests-are-defined-p ()
  (save-excursion
    (goto-char (point-min))
    (go-to-next-commented-test)))

(defun run-commented-tests (&optional beg end)
  (interactive)
  (let ((tests))
    (save-excursion
      (goto-char (or beg (point-min)))
      (while (go-to-next-commented-test end)
	(deftest `(equalp ,(read (match-string 1)) ,(read (match-string 2))) 'auto 'tests)))
    (runtests 'auto tests)))

(defun run-commented-tests-in-region (&optional beg end)
  (interactive "r")
  (run-commented-tests beg end))

(defun tests-wake-up-buttons ()
  (interactive)
  (when (require 'embedded-elisp-library nil t)     
    (save-excursion
      (goto-char (point-min))
      (while (go-to-next-commented-test)
	(save-excursion
	  (beginning-of-line)
	  (eel-wake-up-buttons '(("TEST" . 'run-test-at-point))))))))

(defun run-test-at-point ()
  (interactive)
  (run-commented-tests (point-at-bol) (point-at-eol)))

(defun test-from-region (&optional beg end)
  (interactive "r")
  (let* ((end (min end (point-at-eol)))
         (form (buffer-substring-no-properties beg end))
	 (value (prin1-to-string (eval (read form)))))
    (unless (condition-case nil
		(equalp (eval (read value)) (eval (read form)))
	      (error nil))
      (setq value (concat "'" value))
      (condition-case nil
	  (equalp (eval (read value)) (eval (read form)))
	(error (message "formatting failed ! please check the test"))))
    (kill-region beg end)
    (unless (bolp) (insert ?\n))
    (insert ";TEST " form " => " value)
    (unless (eolp) (insert ?\n))
    (tests-wake-up-buttons)))


;;  misc. utilities 

(defmacro with-temp-buffer-to-string (&rest body) 
  "Evaluate BODY in a temporary buffer then return its contents as a string"
  `(with-temp-buffer
       ,@body
       (buffer-string)))

(put 'with-temp-buffer-to-string 'lisp-indent-function 0)
  
;TEST (with-temp-buffer-to-string (insert "blep"))  => "blep"


;; ===================================================================
;;                 menu
;; ===================================================================

(easy-menu-define tests-menu emacs-lisp-mode-map
  "Menu provided by tests.el for emacs-lisp-mode"
  (let ((treg "^[; \t]*TEST.*=>.*"))
  `("Tests"
    ["Go to next test" go-to-next-commented-test (tests-are-defined-p)]
    ["Go to previous test" go-to-previous-commented-test (tests-are-defined-p)]
    ["Show all tests" (occur ,treg ) (tests-are-defined-p)]
    "--"
    ["Make a test from region" test-from-region (region-exists-p)]
    ["Wake up test buttons" tests-wake-up-buttons (and (tests-are-defined-p)
						   (require 'embedded-elisp-library nil t))]
    "--"
    ["Run test at point" run-test-at-point 
     (string-match ,treg
                   (buffer-substring (point-at-bol) (point-at-eol)))]
    ["Run tests in region"  run-commented-tests-in-region (region-exists-p)]
    ["Run all tests" run-commented-tests (tests-are-defined-p)])))


;; === this is it.
(provide 'tests)

;; tests.el ends here




